# ruff: noqa: N803, N806

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Tuple

import casadi as cs
import numpy as np

from ._type_inference import shape_inference, type_inference

if TYPE_CHECKING:
    from ..typing import ArrayLike, CasadiMatrix, DTypeLike, ShapeLike


DEFAULT_FLOAT = np.float64

SYM_KINDS: Dict[str, Any] = {
    "MX": cs.MX,
    "SX": cs.SX,
    "DM": cs.DM,
    "ndarray": np.ndarray,
}
SYM_NAMES: Dict[Any, str] = {
    cs.MX: "MX",
    cs.SX: "SX",
    cs.DM: "DM",
    np.ndarray: "ndarray",
}

DEFAULT_SYM_TYPE = cs.MX
DEFAULT_SYM_NAME = "MX"


__all__ = [
    "array",
    "sym",
    "sym_like",
    "zeros",
    "ones",
    "zeros_like",
    "ones_like",
]


def _as_casadi_array(x: Any) -> CasadiMatrix | ArrayLike:
    """Convert to a CasADi type (SX, MX, or ndarray)"""
    if isinstance(x, SymbolicArray):
        return x._sym
    # Commented out for codecov... can add back in if these turn out to be necessary
    # elif isinstance(x, (cs.SX, cs.MX)):
    #     return x
    # elif isinstance(x, (tuple, list)):  # Functions with sequence inputs
    #     return [casadi_array(xi) for xi in x]
    else:
        return np.asarray(x)


class SymbolicArray:
    def __init__(
        self,
        sym: cs.SX | cs.MX,
        dtype: DTypeLike | None = None,
        shape: ShapeLike | None = None,
    ):
        # Occasionally CasADi operations will return NotImplemented instead
        # of throwing an error. Ideally we would be able to provide a more
        # helpful error message, but at the very least this should be caught
        # immediately.
        if sym is NotImplemented:
            raise ValueError("SymbolicArray cannot be initialized with NotImplemented")

        if dtype is None:
            dtype = DEFAULT_FLOAT
        self.dtype: np.dtype = np.dtype(dtype)

        if shape is None:
            shape = sym.shape  # type: ignore
        self.shape: ShapeLike = shape

        # Consistent handling of vector shapes
        if len(shape) == 1 and sym.shape[0] == 1:  # type: ignore
            sym = sym.T  # type: ignore
        elif len(shape) == 2:
            sym = cs.reshape(sym, *shape)

        self._sym: cs.SX | cs.MX = sym
        self.kind: str = SYM_NAMES[type(sym)]

    def __array_ufunc__(self, ufunc, method, *inputs, **kwargs):
        """Placeholder for NumPy dispatch (defined in _array_ops)"""
        pass

    def __array_function__(self, func, types, args, kwargs):
        """Placeholder for NumPy dispatch (defined in _array_ops)"""
        pass

    def __repr__(self) -> str:
        return f"{self._sym}"

    def __iter__(self) -> Iterator[SymbolicArray]:
        if len(self.shape) == 0:
            yield self

        else:
            for i in range(self.shape[0]):
                yield self[i]

    def __getitem__(self, index: Any) -> SymbolicArray:
        # This relies on using numpy's indexing and slicing machinery to do
        # shape inference and then applying the same indexing and slicing to
        # the symbolic array.  Probably there will be edge cases where some
        # preprocessing needs to be done on the index before passing it to
        # the symbolic array.  Known issues:
        # - CasADi won't recognize the idiom x[:, None] as a way of adding a
        #   new dimension to x.

        if index is None:
            index = (None,)

        # If a 2D array is indexed with only one index, by default CasADi assumes
        # that the missing index should be 0, whereas NumPy assumes it should be
        # slice(None)
        if len(self.shape) == 2 and not isinstance(index, tuple):
            index = (index, slice(None))

        # Do this before handling the expand_dims cases, because it will make
        # the indices correspond to CasADi rather than numpy.
        result_shape = np.empty(self.shape)[index].shape  # type: ignore

        # Since all CasADi SX objects are 2D, CasADi doesn't recognize the idioms
        # x[:, None] or x[None, :] as a way of adding a new dimension to x, so we
        # need to do it manually.
        if isinstance(index, tuple) and None in index:
            # Cases:
            # - self.shape = () and index = (None,): result_shape = (1,)
            # - self.shape = () and index = (None, None): result_shape = (1, 1)
            # - self.shape = (n,) and index = (None, idx): result_shape = (1, n)
            # - self.shape = (n,) and index = (idx, None): result_shape = (n, 1)
            if self.shape == ():
                index = (slice(None), slice(None))
            elif len(self.shape) == 1:
                # Indexing a (n,) array with (None,) will return a (1, n) array
                if index == (None,):
                    index = (slice(None), slice(None))
                elif index[0] is None:
                    # The underlying symbolic array will have shape (self.shape[0], 1)
                    # so the index needs to be transposed.
                    index = (index[1], slice(None))
                elif index[1] is None:
                    index = (index[0], slice(None))

        return SymbolicArray(self._sym[index], dtype=self.dtype, shape=result_shape)

    def __setitem__(self, index: Any, value: Any) -> None:
        # This assumes that whatever is passed in as `index` will work with
        # CasADi's indexing and slicing machinery.  Probably there will be
        # edge cases where some preprocessing needs to be done on the index
        # before passing it to the underlying symbolic array.
        value = _as_casadi_array(value)  # Make sure it's either SX or ndarray

        # If a 2D array is indexed with only one index, by default CasADi assumes
        # that the missing index should be 0, whereas NumPy assumes it should be
        # slice(None, None, None)
        if len(self.shape) == 2 and not isinstance(index, tuple):
            index = (index, slice(None))

            # Now the problem is that if `value`` is a CasADi SX, then its shape will
            # be (:, index), while `self._sym[index]` will be (index, :).
            value = value.reshape(self._sym[index].shape)  # type: ignore

        self._sym[index] = value

    def __len__(self) -> int:
        if len(self.shape) == 0:
            return 0
        return self.shape[0]

    @property
    def ndim(self) -> int:
        return len(self.shape)

    @property
    def size(self) -> int:
        return int(np.prod(self.shape))

    def __add__(self, other: Any) -> SymbolicArray:
        return np.add(self, other)  # type: ignore

    def __radd__(self, other: Any) -> SymbolicArray:
        return self + other  # type: ignore

    def __sub__(self, other: Any) -> SymbolicArray:
        return np.subtract(self, other)  # type: ignore

    def __rsub__(self, other: Any) -> SymbolicArray:
        return -self + other  # type: ignore

    def __mul__(self, other: Any) -> SymbolicArray:
        return np.multiply(self, other)  # type: ignore

    def __rmul__(self, other: Any) -> SymbolicArray:
        return self * other  # type: ignore

    def __truediv__(self, other: Any) -> SymbolicArray:
        return np.divide(self, other)  # type: ignore

    def __rtruediv__(self, other: Any) -> SymbolicArray:
        return np.divide(other, self)  # type: ignore

    def __pow__(self, other: Any) -> SymbolicArray:
        return np.power(self, other)  # type: ignore

    def __rpow__(self, other: Any) -> SymbolicArray:
        return np.power(other, self)  # type: ignore

    def __mod__(self, other: Any) -> SymbolicArray:
        return np.mod(self, other)  # type: ignore

    def __rmod__(self, other: Any) -> SymbolicArray:
        return np.mod(other, self)  # type: ignore

    def __floordiv__(self, other: Any) -> SymbolicArray:
        return np.floor_divide(self, other)  # type: ignore

    def __rfloordiv__(self, other: Any) -> SymbolicArray:
        return np.floor_divide(other, self)  # type: ignore

    def __divmod__(self, other: Any) -> Tuple[SymbolicArray, SymbolicArray]:
        return np.divmod(self, other)  # type: ignore

    def __rdivmod__(self, other: Any) -> Tuple[SymbolicArray, SymbolicArray]:
        return np.divmod(other, self)  # type: ignore

    def __neg__(self) -> SymbolicArray:
        return np.negative(self)  # type: ignore

    def __abs__(self) -> SymbolicArray:
        return np.fabs(self)  # type: ignore

    def __matmul__(self, other: Any) -> SymbolicArray:
        return np.matmul(self, other)  # type: ignore

    def __gt__(self, other: Any) -> SymbolicArray:
        return np.greater(self, other)  # type: ignore

    def __ge__(self, other: Any) -> SymbolicArray:
        return np.greater_equal(self, other)  # type: ignore

    def __lt__(self, other: Any) -> SymbolicArray:
        return np.less(self, other)  # type: ignore

    def __le__(self, other: Any) -> SymbolicArray:
        return np.less_equal(self, other)  # type: ignore

    def __eq__(self, other: Any) -> SymbolicArray:  # type: ignore
        return np.equal(self, other)  # type: ignore

    def __ne__(self, other: Any) -> SymbolicArray:  # type: ignore
        return np.not_equal(self, other)  # type: ignore

    def __and__(self, other: Any) -> SymbolicArray:
        return np.logical_and(self, other)  # type: ignore

    def __rand__(self, other: Any) -> SymbolicArray:
        return np.logical_and(other, self)  # type: ignore

    def __or__(self, other: Any) -> SymbolicArray:
        return np.logical_or(self, other)  # type: ignore

    def __ror__(self, other: Any) -> SymbolicArray:
        return np.logical_or(other, self)  # type: ignore

    def __xor__(self, other: Any) -> SymbolicArray:
        return np.logical_xor(self, other)  # type: ignore

    def __rxor__(self, other: Any) -> SymbolicArray:
        return np.logical_xor(other, self)  # type: ignore

    def __invert__(self) -> SymbolicArray:
        return np.logical_not(self)  # type: ignore

    @property
    def T(self) -> SymbolicArray:  # noqa: N802
        return np.transpose(self)  # type: ignore

    def simplify(self) -> SymbolicArray:
        return SymbolicArray(cs.simplify(self._sym), dtype=self.dtype, shape=self.shape)

    #
    # Other common NumPy array methods
    #
    def flatten(self, order: str = "C") -> SymbolicArray:
        return np.ravel(self, order=order)  # type: ignore

    def ravel(self, order: str = "C") -> SymbolicArray:
        return np.ravel(self, order=order)  # type: ignore

    def squeeze(self, axis: int | Tuple[int, ...] | None = None) -> SymbolicArray:
        return np.squeeze(self, axis=axis)  # type: ignore

    def reshape(self, shape: ShapeLike, order: str = "C") -> SymbolicArray:
        return np.reshape(self, shape, order=order)  # type: ignore

    def astype(self, dtype: DTypeLike) -> SymbolicArray:
        return np.astype(self, dtype)  # type: ignore

    #
    # Autodiff operations not supported by NumPy
    #
    def grad(self, x: SymbolicArray) -> SymbolicArray:
        dtype = type_inference("default", self, x)
        shape = shape_inference("gradient", self, x)
        return SymbolicArray(cs.gradient(self._sym, x._sym), dtype=dtype, shape=shape)

    def jac(self, x: SymbolicArray) -> SymbolicArray:
        dtype = type_inference("default", self, x)
        shape = shape_inference("jacobian", self, x)
        return SymbolicArray(cs.jacobian(self._sym, x._sym), dtype=dtype, shape=shape)

    def hess(self, x: SymbolicArray) -> SymbolicArray:
        dtype = type_inference("default", self, x)
        shape = shape_inference("hessian", self, x)
        hess, _g = cs.hessian(self._sym, x._sym)
        return SymbolicArray(hess, dtype=dtype, shape=shape)

    def jvp(self, x: SymbolicArray, v: SymbolicArray) -> SymbolicArray:
        dtype = type_inference("default", self, x, v)
        shape = shape_inference("jvp", self, x, v)
        return SymbolicArray(
            cs.jtimes(self._sym, x._sym, v._sym), dtype=dtype, shape=shape
        )

    def vjp(self, x: SymbolicArray, v: SymbolicArray) -> SymbolicArray:
        dtype = type_inference("default", self, x, v)
        shape = shape_inference("vjp", self, x, v)
        return SymbolicArray(
            cs.jtimes(self._sym, x._sym, v._sym, True), dtype=dtype, shape=shape
        )


#
# Factory functions for constructing SymbolicArrays
#


def sym(
    name: str,
    shape: ShapeLike | None = None,
    dtype: DTypeLike = np.float64,
    kind: str = DEFAULT_SYM_NAME,
) -> SymbolicArray:
    """
    Create a symbolic array for use in symbolic computations.

    This function creates symbolic variables that serve as the foundation for building
    computational graphs in Archimedes. Symbolic arrays can be manipulated with
    NumPy-like operations and transformed into efficient computational graphs.

    Parameters
    ----------
    name : str
        Name of the symbolic variable. This name is used for display purposes and
        debugging, and appears in symbolic expression representations.
    shape : int or tuple of int, optional
        Shape of the array. Default is (), which creates a scalar.
        A single integer n creates a vector of length n.
        A tuple (m, n) creates an m×n matrix.
    dtype : numpy.dtype, optional
        Data type of the array. Default is np.float64. This is mostly unused for now
        and all "compiled" functions use float64. It is included for consistency with
        NumPy and will be used for better type control in C code generation in the
        future.
    kind : {"SX", "MX"}, optional
        Kind of symbolic variable to create. Default is "MX".

        - SX: Scalar-based symbolic type. Each element of the array has its own\
        symbolic representation. Generally more efficient for element-wise operations\
        and when computing gradients of scalar functions.

        - MX: Matrix-based symbolic type. The entire array is represented by a single\
        symbolic object. Supports a broader range of operations but may be less\
        efficient for some applications.

    Returns
    -------
    SymbolicArray
        Symbolic array with the given name, shape, dtype, and kind. This object can be
        used in NumPy operations and Archimedes function transformations.

    Notes
    -----
    Symbolic arrays are the building blocks of symbolic computation in Archimedes. They
    represent variables whose values are not specified until later computation,
    allowing for construction of computational graphs that can be efficiently
    evaluated, differentiated, and optimized.

    The difference between SX and MX symbolic types is important:

    - SX types create element-wise symbolic variables and are more efficient for scalar
      operations and gradients of scalar functions.
    - MX types represent entire matrices as single symbolic objects and support a wider
      range of operations, including those that cannot be easily represented
      element-wise.

    Current limitations:

    - Only supports up to 2D arrays (scalars, vectors, and matrices)

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a scalar symbolic variable
    >>> x = arc.sym("x")
    >>> print(x)
    x
    >>>
    >>> # Create a vector (1D array)
    >>> v = arc.sym("v", shape=3)
    >>> print(v)
    v
    >>> print(v.shape)
    (3,)
    >>>
    >>> # Create a matrix (2D array)
    >>> M = arc.sym("M", shape=(2, 2))
    >>> print(M)
    M
    >>> print(M.shape)
    >>> (2, 2)
    >>>
    >>> # Create an SX-type symbolic variable
    >>> y = arc.sym("y", shape=2, kind="SX")
    >>> print(y)
    [y_0, y_1]
    >>>
    >>> # Use in symbolic computations
    >>> f = np.sin(x) + np.cos(x)
    >>> print(f)
    (sin(x)+cos(x))

    See Also
    --------
    array : Create a regular or symbolic array from data
    compile : Create a symbolic function from a Python function
    """
    # TODO: Use `scalar: bool` instead of `kind: str`
    if shape is None:
        shape = ()
    if isinstance(shape, int):
        shape = (shape,)
    if not isinstance(shape, tuple) or not all(isinstance(s, int) for s in shape):
        raise ValueError("Shape must be an int or tuple of ints")
    if not (len(shape) <= 2):
        raise ValueError("Only scalars, vectors, and matrices are supported for now")

    # Note that CasADi creates variables in column-major order, so to be consistent
    # with NumPy we have to reverse the shape and transpose the result.  This only
    # applies to SX, since for MX it's just a single symbol anyway.
    if kind == "SX":
        _sym = cs.SX.sym(name, *shape[::-1]).T
    elif kind == "MX":
        _sym = cs.MX.sym(name, *shape)
    else:
        raise ValueError(f"Unknown symbolic kind {kind}")
    return SymbolicArray(_sym, dtype=dtype, shape=shape)


def sym_like(
    x: ArrayLike,
    name: str,
    dtype: DTypeLike | None = None,
    kind: str = DEFAULT_SYM_NAME,
) -> SymbolicArray:
    """
    Create a symbolic array with the same shape and dtype as an existing array.

    Parameters
    ----------
    x : array_like
        Array to copy shape and dtype from. Can be a NumPy ndarray, SymbolicArray,
        or any array-like object that can be converted to an ndarray.
    name : str
        Name of the symbolic variable. This name is used for display purposes and
        debugging, and appears in symbolic expression representations.
    dtype : numpy.dtype, optional
        Data type of the array. If None (default), uses the dtype of ``x``.
    kind : {"SX", "MX"}, optional
        Kind of symbolic variable to create. Default is "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient for element-wise operations.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Symbolic array with the given name and kind, matching the shape and
        (optionally) dtype of the input array ``x``.

    Notes
    -----
    This function is useful when:

    - Creating symbolic representations of existing numeric data
    - Building symbolic functions that need to match the structure of numeric inputs
    - Prototyping symbolic algorithms using existing arrays as templates

    The function automatically converts non-array inputs to NumPy arrays before
    extracting their shape information.

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Create a symbolic variable with the same shape as a numeric array
    >>> data = np.array([[1.0, 2.0], [3.0, 4.0]])
    >>> M_sym = arc.sym_like(data, "M")
    >>> print(M_sym.shape)
    (2, 2)
    >>>
    >>> # Use with a vector
    >>> v = np.ones(5)
    >>> v_sym = arc.sym_like(v, "v")
    >>> print(v_sym.shape)
    (5,)
    >>>
    >>> # Works with scalar inputs too
    >>> scalar = 42.0
    >>> x = arc.sym_like(scalar, "x")
    >>> print(x.shape)
    ()
    >>>
    >>> # Create a symbolic matrix based on another symbolic array
    >>> original_sym = arc.sym("y", shape=(3, 3))
    >>> another_sym = arc.sym_like(original_sym, "z", kind="SX")
    >>> print(another_sym.shape)
    (3, 3)

    See Also
    --------
    sym : Create a symbolic array with explicit shape
    array : Create a regular or symbolic array from data
    zeros_like : Create an array of zeros with shape/dtype of an input array
    ones_like : Create an array of ones with shape/dtype of an input array
    """
    if not isinstance(x, (np.ndarray, SymbolicArray)):
        x = np.asarray(x)
    return sym(name, x.shape, dtype=dtype or x.dtype, kind=kind)


def array(x: Any, dtype: DTypeLike | None = None) -> ArrayLike:
    """
    Create an array supporting both numeric and symbolic computation.

    This function serves as Archimedes' array creation mechanism, handling both
    numeric and symbolic inputs. It creates the appropriate array type (NumPy or
    SymbolicArray) based on the input data.

    Parameters
    ----------
    x : array_like
        An array-like object, which can be:

        - NumPy ndarray

        - SymbolicArray

        - List or nested list of scalars

        - Scalar value

    dtype : str or numpy.dtype, optional
        The data type for the array. If not specified, dtype is inferred from ``x``.

    Returns
    -------
    array : numpy.ndarray or SymbolicArray
        If the input contains symbolic elements, returns a SymbolicArray.
        Otherwise, returns a NumPy ndarray.

    Notes
    -----
    Array creation using the NumPy dispatch mechanism (``np.array(..., like=...)``) is
    recommended over calling ``array(...)`` directly. The dispatch mechanism supports a
    wider range of input types and better handles numeric input types.

    When working with symbolic computation, this function ensures that array creation
    follows the same patterns as NumPy, allowing for seamless transitions between
    symbolic and numeric computation.

    This function currently supports:

    - Creating arrays from existing arrays (preserving type)
    - Creating arrays from lists of scalars (1D arrays)
    - Creating arrays from lists of lists (2D arrays)

    Limitations and edge cases:

    - Higher-dimensional arrays (>2D) are not supported
    - Creating arrays with inconsistent dimensions will raise a ValueError
    - Complex-valued arrays may have limited symbolic operation support

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # Basic numeric array creation
    >>> arc.array([1.0, 2.0, 3.0])
    array([1., 2., 3.])
    >>>
    >>> # Creating a symbolic array
    >>> x = arc.sym("x", 3)  # Create a symbolic variable
    >>> arc.array([x[0], 2.0, x[2]])  # Mixed symbolic/numeric content
    [x_0, 2, x_2]
    >>>
    >>> # 2D array from nested lists
    >>> arc.array([[1.0, 2.0], [3.0, 4.0]])
    array([[1., 2.],
           [3., 4.]])
    >>>
    >>> # Using NumPy dispatch (recommended approach)
    >>> np.array([x[0], 2.0, x[2]], like=x_sym)  # Creates a SymbolicArray
    [x_0, 2, x_2]

    See Also
    --------
    numpy.array : The NumPy array creation function
    zeros : Create an array of zeros
    ones : Create an array of ones
    sym : Create a symbolic variable
    """

    # Case 1. x is already an array
    if isinstance(x, (SymbolicArray, np.ndarray)):
        if dtype is not None:
            x = x.astype(dtype)
        return x  # type: ignore

    if isinstance(x, (cs.SX, cs.MX)):
        return SymbolicArray(x, dtype=dtype)

    if (np.isscalar(x) and np.isreal(x)) or isinstance(x, cs.DM):
        return np.array(x, dtype=dtype)

    return _dispatch_array(x, dtype=dtype)


def _empty_like(x: ArrayLike) -> np.ndarray:
    if isinstance(x, SymbolicArray):
        return np.empty(x.shape, dtype=x.dtype)
    else:
        return x  # type: ignore


def _result_type(*inputs: Any) -> np.dtype:
    np_inputs = tuple(map(_empty_like, inputs))
    return np.result_type(*np_inputs)  # type: ignore


def _is_list_of_scalars(x: List[Any]) -> bool:
    return all(isinstance(xi, (int, float)) or len(xi) <= 1 for xi in x)  # type: ignore


def _dispatch_array(x: Any, dtype: DTypeLike | None = None) -> ArrayLike:
    """`array` function dispatched from np.array(..., like=[SymbolicArray])"""
    # For now, support three cases:
    # 1. x is already an array (handled by main array function)
    # 2. x is a list of scalars (convert to an (n,) array)
    # 3. x is a list of lists (convert to an (n, m) array)

    # Case 1: x is already an array
    # This is handled by `array`, so we don't need to do anything here.
    # Confirmed by code coverage analysis

    if isinstance(x, (list, tuple)):
        if len(x) == 0:
            return np.array(x, dtype=dtype)

        # Case 2. x is a list of scalars
        if not all(
            isinstance(xi, (list, tuple, np.ndarray, SymbolicArray)) for xi in x
        ):
            # Check that everything is a scalar
            if not _is_list_of_scalars(x):  # type: ignore
                raise ValueError(f"Creating array with inconsistent data: {x}")
            result_shape = (len(x),)
            result_dtype = dtype or _result_type(*x)
            cs_x = list(map(_as_casadi_array, x))
            arr = cs.vcat(cs_x)  # type: ignore
            # if isinstance(arr, cs.DM):
            #     return np.array(arr, dtype=result_dtype).reshape(result_shape)
            return SymbolicArray(arr, dtype=result_dtype, shape=result_shape)

        # Case 3. x is a list of lists, arrays, etc
        # check that all lists are only scalars
        if not all(
            _is_list_of_scalars(xi)
            for xi in x  # type: ignore
        ):
            raise ValueError(
                "Can only create array from list of scalars or list of lists of scalars"
            )

        # Check that the lengths are consistent
        len_set = set(map(len, x))
        if len(len_set) != 1:
            raise ValueError(f"Inconsistent lengths in list of lists: {len_set}")

        result_shape = (len(x),) if len(x[0]) == 0 else (len(x), len(x[0]))  # type: ignore
        result_dtype = dtype or _result_type(*[_result_type(*xi) for xi in x])  # type: ignore
        cs_x = [list(map(_as_casadi_array, xi)) for xi in x]

        arr = cs.vcat([cs.hcat(xi) for xi in cs_x])  # type: ignore
        if isinstance(arr, cs.DM):
            return np.array(arr, dtype=result_dtype).reshape(result_shape)
        return SymbolicArray(
            arr,
            dtype=result_dtype,
            shape=result_shape,
        )

    raise NotImplementedError(
        f"Converting {x} (type={type(x)}) to array is not supported"
    )


def _np_shape(shape: ShapeLike) -> ShapeLike:
    # Check that the shape is valid and return a tuple of ints.
    if isinstance(shape, int):
        return (shape,)
    if not isinstance(shape, tuple):
        raise ValueError("shape must be an int or a tuple of ints")
    if not all(isinstance(s, int) for s in shape):
        raise ValueError("shape must be a tuple of ints")
    return shape


# zeros, ones, zeros_like, eye, diag
def _cs_shape(shape: ShapeLike) -> Tuple[int, int]:
    # The shape of the CasADi object is always 2D, so we need to handle the
    # cases where the specified shape is () or (n,) separately.
    cs_shape = _np_shape(shape)

    if len(cs_shape) > 2:
        raise ValueError("Only scalars, vectors, and matrices are supported for now")

    if len(cs_shape) == 0:
        cs_shape = (1, 1)  # type: ignore

    elif len(cs_shape) == 1:
        cs_shape = (cs_shape[0], 1)  # type: ignore

    return cs_shape  # type: ignore


def zeros(
    shape: ShapeLike,
    dtype: DTypeLike = np.float64,
    sparse: bool = True,
    kind: str = DEFAULT_SYM_NAME,
) -> SymbolicArray:
    """
    Construct a symbolic array of zeros with the given shape and dtype.

    This function creates an array filled with zeros, either as "structural"
    (sparse) zeros that are symbolic placeholders, or as actual numerical
    zero values.

    Parameters
    ----------
    shape : int or tuple of ints
        Shape of the array. A single integer n creates a vector of length n.
        A tuple (m, n) creates an m×n matrix.
    dtype : numpy.dtype, optional
        Data type of the array. Default is np.float64.
    sparse : bool, optional
        If True (default), the array will contain "structural" zeros, which are
        symbolic placeholders rather than actual numeric values. These are more
        efficient in memory and computation.
        If False, the array will contain actual numeric zero values.
    kind : {"SX", "MX"}, optional
        Kind of symbolic variable to create. Default is "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient for element-wise operations.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Symbolic array of zeros with the given shape, dtype, and symbolic kind.

    Notes
    -----
    The distinction between sparse (structural) and dense (numerical) zeros is
    important:

    - Sparse/structural zeros (sparse=True) are efficient for building computational
      graphs where many elements are zero, especially in matrix operations.

    - Dense/numerical zeros (sparse=False) actually contain the value 0 and behave
      more like traditional NumPy arrays filled with zeros.

    When working within a function that will be executed with both symbolic and
    numeric arrays, prefer using ``np.zeros_like`` or ``np.zeros(..., like=x)`` where
    ``x`` is either a SymbolicArray or NumPy array. This provides better compatibility
    across both numeric and symbolic execution paths.

    The exception is when you specifically need sparse/structural zeros, which are only
    available through this direct function call with sparse=True.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a vector of structural zeros (sparse representation)
    >>> z1 = arc.zeros(5)
    >>> print(z1)
    [00, 00, 00, 00, 00]
    >>>
    >>> # Create a matrix of numerical zeros (dense representation)
    >>> z2 = arc.zeros(5, sparse=False)
    >>> print(z2)
    @1=0, [@1, @1, @1, @1, @1]
    >>>
    >>> # Create MX-type zeros
    >>> z3 = arc.zeros(4, kind="MX")

    See Also
    --------
    numpy.zeros : NumPy's array of zeros function
    zeros_like : Create an array of zeros with shape/dtype of an input array
    ones : Create an array of ones
    array : Create an array from data
    """
    sym_typ = SYM_KINDS[kind]
    _zeros = sym_typ if sparse else sym_typ.zeros
    return SymbolicArray(_zeros(*_cs_shape(shape)), dtype=dtype, shape=_np_shape(shape))


def ones(
    shape: ShapeLike, dtype: DTypeLike = np.float64, kind: str = "MX"
) -> SymbolicArray:
    """
    Construct a symbolic array of ones with the given shape and dtype.

    This function creates an array filled with the value 1, equivalent to
    NumPy's ones function but returning a symbolic array suitable for use
    in symbolic computations.

    Parameters
    ----------
    shape : int or tuple of ints
        Shape of the array. A single integer n creates a vector of length n.
        A tuple (m, n) creates an m×n matrix.
    dtype : numpy.dtype, optional
        Data type of the array. Default is np.float64.
    kind : {"SX", "MX"}, optional
        Kind of symbolic variable to create. Default is "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient for element-wise operations.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Symbolic array of ones with the given shape, dtype, and symbolic kind.

    Notes
    -----
    This function is the symbolic counterpart to NumPy's `ones` function. It's useful
    for initializing arrays with ones in symbolic computation contexts, such as creating
    weight matrices, mask arrays, or default values for symbolic computation.

    When working within a function that will be executed with both symbolic and
    numeric arrays, prefer using ``np.ones_like`` or ``np.ones(..., like=x)`` where
    ``x`` is either a SymbolicArray or NumPy array. This provides better compatibility
    across both numeric and symbolic execution paths.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a vector of ones
    >>> o1 = arc.ones(5)
    >>> print(o1)
    [1, 1, 1, 1, 1]
    >>>
    >>> # Create a matrix of ones
    >>> o2 = arc.ones((2, 3))
    >>> print(o2)
    [[1, 1, 1], [1, 1, 1]]
    >>>
    >>> # Create MX-type ones
    >>> o3 = arc.ones(4, kind="MX")

    See Also
    --------
    numpy.ones : NumPy's array of ones function
    ones_like : Create an array of ones with shape/dtype of an input array
    zeros : Create an array of zeros
    eye : Create an identity matrix
    array : Create an array from data
    """
    return SymbolicArray(
        SYM_KINDS[kind].ones(*_cs_shape(shape)), dtype=dtype, shape=_np_shape(shape)
    )


def zeros_like(
    x: ArrayLike,
    dtype: DTypeLike | None = None,
    sparse: bool = True,
    kind: str | None = None,
) -> SymbolicArray:
    """
    Create a symbolic array of zeros with the same shape and dtype as an input array.

    This function constructs a symbolic array filled with zeros, matching the dimensions
    and data type of an existing array. It's useful for creating compatible zero arrays
    in functions that need to work with both symbolic and numeric inputs.

    Parameters
    ----------
    x : array_like
        The array whose shape and dtype will be used. Can be a NumPy ndarray,
        SymbolicArray, or any array-like object that can be converted to an array.
    dtype : numpy.dtype, optional
        Data type of the new array. If None (default), uses the dtype of ``x``.
    sparse : bool, optional
        If True (default), creates "structural" zeros, which are symbolic
        placeholders rather than actual numeric values. These are more efficient
        for building computational graphs.
        If False, creates numerical zero values.
    kind : {"SX", "MX"} or None, optional
        Kind of symbolic variable to create. If None (default), uses the kind
        of ``x`` if it's a SymbolicArray, otherwise uses "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient for element-wise operations.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Symbolic array of zeros with the same shape as ``x``, and with the
        specified dtype and symbolic kind.

    Notes
    -----
    This function is the symbolic counterpart to NumPy's :py:func:`np.zeros_like`.
    While creating a standard NumPy array of zeros requires numeric inputs, this
    function works with both symbolic and numeric arrays, preserving the symbolic
    nature when needed.

    When used inside a function decorated with :py:func:`compile`, this function helps
    create arrays that match the input's shape, which is useful for maintaining
    compatibility between symbolic and numeric execution paths.

    The ``sparse`` parameter determines whether the zeros are "structural" (symbolic
    placeholders) or actual numeric zeros, which can affect memory usage and
    computational efficiency.

    When the array is not sparse, it is preferred to use ``np.zeros_like`` or
    ``np.zeros(..., like=x)``, where ``x`` is either a SymbolicArray or NumPy array.
    This can provide better compatibility across both numeric and symbolic execution
    paths. The exception is when you specifically need sparse/structural zeros, which
    are only available through this direct function call with ``sparse=True``.

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # With a symbolic input array
    >>> x_sym = arc.sym("x", shape=(2, 3))
    >>> z_sym = arc.zeros_like(x_sym)
    >>> print(z_sym.shape)
    (2, 3)
    >>>
    >>> # With a numeric input array
    >>> x_num = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
    >>> z_num = arc.zeros_like(x_num)
    >>> print(z_num.shape)
    (3, 2)
    >>>
    >>> # Creating numerical zeros instead of structural zeros
    >>> z_dense = arc.zeros_like(x_sym, sparse=False)
    >>>
    >>> # Changing the kind of symbolic variable
    >>> z_mx = arc.zeros_like(x_sym, kind="MX")
    >>>
    >>> # In a function that will be traced symbolically:
    >>> @arc.compile
    >>> def process_array(x):
    >>>     # Create a result array with same shape as input
    >>>     # Dispatches to this function when x is a SymbolicArray
    >>>     result = np.zeros_like(x)
    >>>     for i in range(x.shape[0]):
    >>>         result[i] = x[i] ** 2
    >>>     return result

    See Also
    --------
    numpy.zeros_like : NumPy's equivalent function for numeric arrays
    zeros : Create a symbolic array of zeros with specified shape
    ones_like : Create a symbolic array of ones with same shape as input
    array : Create an array from data
    """
    if kind is None:
        if isinstance(x, SymbolicArray):
            kind = x.kind
        else:
            kind = DEFAULT_SYM_NAME
    x = array(x)  # Should be SymbolicArray or ndarray
    return zeros(x.shape, dtype=dtype or x.dtype, sparse=sparse, kind=kind)


def ones_like(
    x: ArrayLike, dtype: DTypeLike | None = None, kind: str | None = None
) -> SymbolicArray:
    """
    Create a symbolic array of ones with the same shape and dtype as an input array.

    This function constructs a symbolic array filled with ones, matching the dimensions
    and data type of an existing array. It's useful for creating masks, weights, or
    initial values in functions that need to work with both symbolic and numeric inputs.

    Parameters
    ----------
    x : array_like
        The array whose shape and dtype will be used. Can be a NumPy ndarray,
        SymbolicArray, or any array-like object that can be converted to an array.
    dtype : numpy.dtype, optional
        Data type of the new array. If None (default), uses the dtype of ``x``.
    kind : {"SX", "MX"} or None, optional
        Kind of symbolic variable to create. If None (default), uses the kind
        of `x` if it's a SymbolicArray, otherwise uses "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient for element-wise operations.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Symbolic array of ones with the same shape as ``x``, and with the
        specified dtype and symbolic kind.

    Notes
    -----
    This function is the symbolic counterpart to NumPy's `ones_like`. While a standard
    NumPy array of ones can only hold numeric values, this function works with both
    symbolic and numeric entries, preserving the symbolic nature when needed.

    When used inside a function decorated with :py:func:``compile``, this function
    helps create arrays of ones that match the input's shape, which is useful for
    for initializing accumulators, creating masks, or setting default values.

    Examples
    --------
    >>> import numpy as np
    >>> import archimedes as arc
    >>>
    >>> # With a symbolic input array
    >>> x_sym = arc.sym("x", shape=(2, 3), kind="SX")
    >>> o_sym = arc.ones_like(x_sym)
    >>> print(o_sym.shape)
    (2, 3)
    >>>
    >>> # With a numeric input array
    >>> x_num = np.array([[1.0, 2.0], [3.0, 4.0], [5.0, 6.0]])
    >>> o_num = arc.ones_like(x_num)
    >>> print(o_num.shape)
    (3, 2)
    >>>
    >>> # Changing the kind of symbolic variable
    >>> o_mx = arc.ones_like(x_sym, kind="MX")
    >>>
    >>> # In a function that will be traced symbolically:
    >>> @arc.compile
    >>> def process_array(x):
    >>>     # Initialize result with ones. Dispatches to this function when x is a
    >>>     # SymbolicArray
    >>>     result = np.ones_like(x)
    >>>     for i in range(x.shape[0]):
    >>>         result[i] *= x[i]
    >>>     return result

    See Also
    --------
    numpy.ones_like : NumPy's equivalent function for numeric arrays
    ones : Create a symbolic array of ones with specified shape
    zeros_like : Create a symbolic array of zeros with same shape as input
    empty_like : Create an uninitialized array with same shape as input
    """
    x = array(x)  # Should be SymbolicArray or ndarray
    if kind is None:
        if isinstance(x, SymbolicArray):
            kind = x.kind
        else:
            kind = DEFAULT_SYM_NAME
    return ones(x.shape, dtype=dtype or x.dtype, kind=kind)


def eye(
    N: int,
    M: int | None = None,
    k: int = 0,
    dtype: DTypeLike = np.float64,
    order: str = "C",
    device: str = "cpu",
    kind: str = DEFAULT_SYM_NAME,
) -> SymbolicArray:
    """
    Construct a symbolic identity matrix of size `n` with the given dtype.

    This function creates an n×n matrix with ones on the diagonal and zeros elsewhere,
    equivalent to NumPy's eye function but returning a symbolic array suitable for
    use in symbolic computations.

    Parameters
    ----------
    N : int
        Size of the identity matrix.
    M : int, optional
        Number of columns in the identity matrix. If None (default), M is set to N,
        creating a square matrix.
    k : int, optional
        Index of the diagonal. If None (default), the main diagonal is used.
    dtype : numpy.dtype, optional
        Data type of the array. Default is np.float64.
    order : {"C", "F"}, optional
        Memory layout order of the array. For compatibility with NumPy only - ignored
        by this implementation.
    device: str, optional
        Device to create the array on. For compatibility with NumPy only - ignored
        by this implementation.
    kind : {"SX", "MX"}, optional
        Kind of symbolic variable to create. Default is "MX".

        - SX: Scalar-based symbolic type. Each element has its own symbolic\
        representation. Generally more efficient.

        - MX: Matrix-based symbolic type. The entire array is represented by\
        a single symbolic object. Supports a broader range of operations.

    Returns
    -------
    SymbolicArray
        Identity matrix of shape (n, n) with the given dtype and symbolic kind.

    Notes
    -----
    The identity matrix is a square matrix with ones on the main diagonal and zeros
    elsewhere.

    When working within a function that will be executed with both symbolic and
    numeric arrays, prefer using `np.eye(..., like=x)` where `x` is either a
    SymbolicArray or NumPy array, for better compatibility with numerical inputs.
    This function will automatically be dispatched to by NumPy when `x` is symbolic.

    Examples
    --------
    >>> import archimedes as arc
    >>> import numpy as np
    >>>
    >>> # Create a 3×3 identity matrix
    >>> I = arc.eye(3, kind="SX")
    >>> print(I)
    [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    >>>
    >>> # Create an MX-type identity matrix (default)
    >>> I_mx = arc.eye(4)
    >>>
    >>> # Use in matrix operations
    >>> x = arc.sym("x", shape=(3, 3))
    >>> identity_transform = I @ x  # Equivalent to x

    See Also
    --------
    numpy.eye : NumPy's identity matrix function
    ones : Create an array of ones
    zeros : Create an array of zeros
    """
    if M is None:
        M = N

    if k != 0:
        raise NotImplementedError(
            "Creating identity matrices with a non-zero diagonal (k != 0) "
            "is not supported yet"
        )

    sz = max(N, M)
    eye_mat = SymbolicArray(SYM_KINDS[kind].eye(sz), dtype=dtype, shape=(sz, sz))

    eye_mat = eye_mat[:N, :M]  # Ensure it has shape (N, M)
    return eye_mat



# ruff: noqa: N806

from __future__ import annotations

import logging
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Callable, Tuple, TypeVar

import numpy as np
import osqp
from scipy import sparse
from scipy.optimize import OptimizeResult

import archimedes as arc
from archimedes import tree

from ._common import _ravel_args

if TYPE_CHECKING:
    from archimedes.typing import Tree

    T = TypeVar("T", bound=Tree)


__all__ = [
    "lm_solve",
    "LMStatus",
    "LMResult",
]


class LMStatus(IntEnum):
    """Status codes for Levenberg-Marquardt optimization results.

    These codes follow the MINPACK convention and provide detailed information
    about the termination condition of the optimization algorithm. Understanding
    these codes is essential for interpreting optimization results and diagnosing
    potential issues.

    Attributes
    ----------
    FTOL_REACHED : int = 1
        Both actual and predicted relative reductions in the objective function
        are at most ``ftol``. This indicates convergence based on minimal
        improvement in the objective function.

    XTOL_REACHED : int = 2
        Relative error between two consecutive parameter iterates is at most
        ``xtol``. This indicates convergence based on minimal change in the
        parameters.

    BOTH_TOL_REACHED : int = 3
        Both ``ftol`` and ``xtol`` conditions are satisfied simultaneously.
        This represents the strongest convergence criterion.

    GTOL_REACHED : int = 4
        The cosine of the angle between the gradient and any column of the
        Jacobian is at most ``gtol`` in absolute value. For constrained
        problems, this applies to the projected gradient. This indicates
        convergence to a critical point.

    MAX_FEVAL : int = 5
        Maximum number of function evaluations (``max_nfev``) has been reached
        without achieving convergence. This typically indicates that either
        more iterations are needed, the tolerances are too tight, or the
        problem is ill-conditioned.

    Notes
    -----
    **Success vs. Failure**:
        Status codes 1-4 indicate successful convergence, while code 5 indicates
        failure to converge within the iteration limit. Use the ``success``
        property to check for overall success.

    **Interpretation Guide**:
        - **FTOL_REACHED**: Good for most applications, indicates objective
          function has stopped improving significantly
        - **XTOL_REACHED**: May indicate convergence to a local minimum or
          numerical precision limits
        - **BOTH_TOL_REACHED**: Strongest convergence guarantee
        - **GTOL_REACHED**: Gradient-based convergence, good indicator of
          optimality conditions
        - **MAX_FEVAL**: Check if more iterations are needed or if the problem
          requires reformulation
    """

    # Success codes (convergence achieved)
    FTOL_REACHED = 1  # Function tolerance convergence
    XTOL_REACHED = 2  # Parameter tolerance convergence
    BOTH_TOL_REACHED = 3  # Both ftol and xtol satisfied
    GTOL_REACHED = 4  # Gradient tolerance convergence

    # Failure codes
    MAX_FEVAL = 5  # Maximum function evaluations reached

    @property
    def message(self) -> str:
        """Get descriptive message for this status code."""
        messages = {
            self.FTOL_REACHED: (
                "Both actual and predicted relative reductions in the sum of squares "
                "are at most ftol"
            ),
            self.XTOL_REACHED: (
                "Relative error between two consecutive iterates is at most xtol"
            ),
            self.BOTH_TOL_REACHED: "Conditions for ftol and xtol both hold",
            self.GTOL_REACHED: (
                "The cosine of the angle between fvec and any column of the Jacobian "
                "is at most gtol in absolute value"
            ),
            self.MAX_FEVAL: "Number of function evaluations has reached max_nfev",
        }
        return messages.get(self, "Unknown status")

    @property
    def success(self) -> bool:
        """Check if this status indicates successful convergence."""
        return self in (
            self.FTOL_REACHED,
            self.XTOL_REACHED,
            self.BOTH_TOL_REACHED,
            self.GTOL_REACHED,
        )


class LMProgress:
    """Handle progress reporting for LM optimization."""

    def __init__(self, logger=None):
        self.logger = logger
        self.iteration = 0
        self.prev_cost = None
        self.header_printed = False

    def report(self, cost, grad_norm, step_norm, nfev):
        """Report progress in SciPy-style format."""
        if not self.header_printed:
            self._print_header()
            self.header_printed = True

        # Calculate cost reduction
        if self.prev_cost is not None:
            cost_reduction = self.prev_cost - cost
        else:
            cost_reduction = None

        self._print_iteration(
            self.iteration, nfev, cost, cost_reduction, step_norm, grad_norm
        )

        self.prev_cost = cost
        self.iteration += 1

    def _print_header(self):
        msg = (
            f"{'Iteration':^10} {'Total nfev':^12} {'Cost':^15} "
            f"{'Cost reduction':^15} {'Step norm':^12} {'Optimality':^12}"
        )
        if self.logger is None:
            print(msg)
        else:
            self.logger.info(msg)

    def _print_iteration(
        self, iter_num, nfev, cost, cost_reduction, step_norm, grad_norm
    ):
        """Print a single iteration row."""
        # Format numbers with appropriate precision
        cost_str = f"{cost:.4e}"
        grad_str = f"{grad_norm:.2e}"

        if cost_reduction is not None:
            cost_red_str = f"{cost_reduction:.2e}"
            step_str = f"{step_norm:.2e}" if step_norm is not None else ""
        else:
            cost_red_str = ""
            step_str = ""

        msg = (
            f"{iter_num:^10} {nfev:^12} {cost_str:^15} "
            f"{cost_red_str:^15} {step_str:^12} {grad_str:^12}"
        )
        if self.logger is None:
            print(msg)
        else:
            self.logger.info(msg)


class LMResult(OptimizeResult):
    """Result of Levenberg-Marquardt optimization.

    This class provides detailed information about the optimization process
    and results, following SciPy conventions while adding specialized fields
    for system identification applications.

    Attributes
    ----------
    x : ndarray or Tree
        Solution parameters with the same structure as the initial guess.
        For system identification, this preserves the nested parameter
        organization (e.g., ``{"mass": 1.0, "damping": {"c1": 0.1}}``).
    success : bool
        Whether optimization terminated successfully. True for status codes
        1-4 (converged), False for status code 5 (max iterations reached).
    status : LMStatus
        Detailed termination status code indicating the specific convergence
        criterion that was satisfied or reason for termination.
    message : str
        Human-readable description of the termination reason corresponding
        to the status code.
    fun : float
        Final objective function value. For least-squares problems, this
        is 0.5 times the sum of squared residuals.
    jac : ndarray
        Final gradient vector of shape ``(n,)``. For constrained problems,
        this is the full gradient (not projected).
    hess : ndarray
        Final Hessian matrix or approximation of shape ``(n, n)``. This
        can be used for uncertainty quantification and further analysis.
    nfev : int
        Number of objective function evaluations performed during optimization.
        Each evaluation computes the objective value, gradient, and Hessian.
    njev : int
        Number of Jacobian evaluations. For this implementation, this equals
        ``nfev`` since gradient and Hessian are computed simultaneously.
    nit : int
        Number of algorithm iterations. Each iteration may involve multiple
        function evaluations due to the trust region approach.
    history : List[Dict[str, Any]]
        Detailed iteration history containing convergence diagnostics:

        - ``iter`` : Iteration number
        - ``cost`` : Objective function value
        - ``grad_norm`` : Gradient norm (or projected gradient for constrained)
        - ``lambda`` : Levenberg-Marquardt damping parameter
        - ``x`` : Parameter values at this iteration
        - ``step_norm`` : Step size (when step is accepted)
        - ``actred`` : Actual reduction in objective
        - ``prered`` : Predicted reduction from quadratic model
        - ``ratio`` : Ratio of actual to predicted reduction

    See Also
    --------
    lm_solve : Function that returns this result type
    LMStatus : Detailed description of status codes
    """

    # x: np.ndarray
    # success: bool
    # status: LMStatus
    # message: str
    # fun: float
    # jac: np.ndarray
    # hess: np.ndarray
    # nfev: int
    # njev: int
    # nit: int
    # history: List[Dict[str, Any]]


def _compute_step(hess, grad, diag, lambda_val, x, bounds=None, qp_solver=None):
    """
    Compute the Levenberg-Marquardt step by solving the damped normal equations.

    Uses hybrid approach: first try unconstrained step, then QP if bounds violated.

    Parameters
    ----------
    hess : ndarray, shape (n, n)
        Hessian matrix or approximation
    grad : ndarray, shape (n,)
        Gradient vector
    diag : ndarray, shape (n,)
        Scaling factors for the variables
    lambda_val : float
        Levenberg-Marquardt damping parameter
    x : ndarray, shape (n,)
        Current parameter values
    bounds : tuple, optional
        Tuple of (lower_bounds, upper_bounds) for parameters.
    qp_solver : osqp.OSQP, optional
        OSQP solver instance for handling bounds. Must be provided if bounds are set.

    Returns
    -------
    step : ndarray, shape (n,)
        Step direction
    """
    n = len(grad)
    # Form the damped Hessian: H + λ·diag(diag)²
    H_damped = hess.copy()
    for i in range(n):
        H_damped[i, i] += lambda_val * diag[i] ** 2

    # Step 1: Compute unconstrained step
    step = None
    try:
        # Use Cholesky decomposition for better numerical stability
        # This requires the matrix to be positive definite
        L = np.linalg.cholesky(H_damped)  # H_damped = L @ L.T
        # First solve L @ y = -grad
        y = np.linalg.solve(L, -grad)
        # Then solve L.T @ step = y
        step = np.linalg.solve(L.T, y)
    except np.linalg.LinAlgError:
        # Fallback for ill-conditioned or non-positive definite matrices
        try:
            # Try standard solver
            step = np.linalg.solve(H_damped, -grad)
        except np.linalg.LinAlgError:
            # Last resort: gradient descent direction with normalization
            step = -grad / (np.linalg.norm(grad) + 1e-8)

    # Step 2: Check if bounds are satisfied
    if bounds is None:
        return step

    lb, ub = bounds
    x_trial = x + step

    # Check if trial point violates bounds
    violates_bounds = np.any(x_trial < lb) or np.any(x_trial > ub)

    if not violates_bounds:
        return step  # Fast path: no QP needed

    # Step 3: Solve constrained QP
    if qp_solver is None:
        raise ValueError("qp_solver must be provided if bounds are set.")

    # Update QP matrices: min 0.5*p^T*H_damped*p + grad^T*p
    # subject to: lb <= x + p <= ub
    # Rearranged: (lb - x) <= p <= (ub - x)
    l_qp = lb - x  # Lower bounds for step
    u_qp = ub - x  # Upper bounds for step

    # Update the QP problem
    qp_solver.update(P=H_damped, q=grad, l=l_qp, u=u_qp)

    # Warm start with unconstrained solution (projected to bounds)
    qp_solver.warm_start(x=np.clip(step, l_qp, u_qp))

    # Solve QP
    qp_results = qp_solver.solve()

    if qp_results.info.status != "solved":
        print(f"OSQP solver status: {qp_results.info.status}")
        # Fallback: project unconstrained step to bounds
        return np.clip(step, l_qp, u_qp)

    return qp_results.x


def _compute_predicted_reduction(grad, step, hess, current_objective=None):
    """
    Compute the predicted reduction in the objective function.

    For the quadratic model q(p) = f + g^T·p + 0.5·p^T·H·p,
    the predicted reduction is: pred_red = -(g^T·p + 0.5·p^T·H·p)

    Parameters
    ----------
    grad : ndarray, shape (n,)
        Gradient at the current point
    step : ndarray, shape (n,)
        Proposed step
    hess : ndarray, shape (n, n)
        Hessian matrix or approximation
    current_objective : float, optional
        Current objective function value for scaling (if provided)

    Returns
    -------
    pred_red : float
        Predicted reduction in the objective function
    """
    # For step computed from (H + λI)p = -g, we expect:
    # pred_red = -g^T·p - 0.5·p^T·H·p
    linear_term = np.dot(grad, step)
    quadratic_term = 0.5 * np.dot(step, hess @ step)
    pred_red = -(linear_term + quadratic_term)

    # Scale by current objective to make it relative
    # (like MINPACK does with residual norm)
    if current_objective is not None and current_objective > 0:
        # Add small epsilon to prevent division by zero near optimum
        epsilon = 1e-16
        pred_red = pred_red / (current_objective + epsilon)

    return pred_red


def _project_gradient(grad, x, bounds, atol=1e-8):
    """
    Project gradient onto the tangent cone of box constraints.

    This computes the "projected gradient" which is the correct measure
    for convergence in constrained optimization. At a constrained optimum,
    the projected gradient should be zero, not the full gradient.

    Parameters
    ----------
    grad : ndarray, shape (n,)
        Current gradient vector
    x : ndarray, shape (n,)
        Current parameter values
    bounds : tuple of (lower, upper)
        Box constraints where lower and upper are arrays of shape (n,)
        Use -np.inf/np.inf for unbounded variables
    atol : float, optional
        Absolute tolerance for determining if a variable is at its bound

    Returns
    -------
    grad_proj : ndarray, shape (n,)
        Projected gradient
    active_lower : ndarray, shape (n,), dtype=bool
        True where lower bounds are active
    active_upper : ndarray, shape (n,), dtype=bool
        True where upper bounds are active

    Mathematical Notes
    -----------------
    For box constraints l ≤ x ≤ u, the projected gradient is:

    g_proj[i] = {
        0       if x[i] = l[i] and g[i] > 0  (at lower bound, gradient points out)
        0       if x[i] = u[i] and g[i] < 0  (at upper bound, gradient points out)
        g[i]    otherwise                    (interior or admissible direction)
    }

    This captures the KKT optimality conditions:
    - Interior variables: g_proj[i] = g[i] = 0 at optimum
    - Boundary variables: g_proj[i] = 0 always (projected out)
    """
    lower, upper = bounds

    # Determine which constraints are active (within tolerance)
    active_lower = (x <= lower + atol) & np.isfinite(lower)
    active_upper = (x >= upper - atol) & np.isfinite(upper)

    # Start with full gradient
    grad_proj = grad.copy()

    # Zero out gradient components that point "outward" from active constraints
    # At lower bound: zero positive gradients (can't move further left)
    grad_proj = np.where(active_lower & (grad > 0), 0.0, grad_proj)

    # At upper bound: zero negative gradients (can't move further right)
    grad_proj = np.where(active_upper & (grad < 0), 0.0, grad_proj)

    return grad_proj, active_lower, active_upper


def _check_constrained_convergence(grad, x, bounds, gtol=1e-8, atol=1e-8):
    """
    Check convergence for box-constrained optimization using projected gradient.

    Parameters
    ----------
    grad : ndarray
        Current gradient
    x : ndarray
        Current parameters
    bounds : tuple of (lower, upper)
        Box constraints
    gtol : float
        Gradient tolerance for convergence
    atol : float
        Tolerance for determining active constraints

    Returns
    -------
    converged : bool
        True if converged according to constrained optimality conditions
    grad_proj_norm : float
        Norm of projected gradient (should be small at optimum)
    active_info : dict
        Information about active constraints
    """
    grad_proj, active_lower, active_upper = _project_gradient(grad, x, bounds, atol)

    grad_proj_norm = np.linalg.norm(grad_proj, np.inf)
    converged = grad_proj_norm <= gtol

    active_info = {
        "n_active_lower": np.sum(active_lower),
        "n_active_upper": np.sum(active_upper),
        "total_active": np.sum(active_lower | active_upper),
        "grad_proj_norm": grad_proj_norm,
        "full_grad_norm": np.linalg.norm(grad, np.inf),
    }

    return converged, grad_proj_norm, active_info


# TODO:
# - support sparse matrices for Hessian and Jacobian
def lm_solve(
    func: Callable,
    x0: T,
    args: Tuple[Any, ...] = (),
    bounds: Tuple[T, T] | None = None,
    ftol: float = 1e-6,
    xtol: float = 1e-6,
    gtol: float = 1e-6,
    max_nfev: int = 100,
    diag: T | None = None,
    lambda0: float = 1e-3,
    log_level: int | None = None,
) -> OptimizeResult:
    """Solve nonlinear least squares using modified Levenberg-Marquardt algorithm.

    Solves optimization problems of the form:

    .. code-block:: text

        minimize    0.5 * ||r(x)||²
        subject to  lb <= x <= ub

    where ``r(x)`` are residuals computed by the objective function.

    This implementation uses a direct Hessian approximation and Cholesky
    factorization instead of the standard QR approach.  It also supports
    box constraints by switching to a quadratic programming solver
    ([OSQP](https://osqp.org/)) when bounds are active.

    Parameters
    ----------
    func : callable
        Objective function with signature ``func(x, *args) -> r``, where ``r`` is
        a vector of residuals.
    x0 : Tree
        Initial parameter guess. Can be a flat array or a tree structure which
        will be preserved in the solution.
    args : tuple, optional
        Extra arguments passed to the objective function.
    bounds : tuple of (Tree, Tree), optional
        Box constraints specified as ``(lower_bounds, upper_bounds)``.
        Each bound array must have the same tree structure as ``x0``. Use
        ``-np.inf`` and ``np.inf`` for unbounded variables. Enables physical
        constraints like mass > 0, damping > 0, etc.
    ftol : float, default=1e-6
        Tolerance for relative reduction in objective function.
        Convergence occurs when both actual and predicted reductions
        are smaller than this value.
    xtol : float, default=1e-6
        Tolerance for relative change in parameters. Convergence occurs
        when the relative step size is smaller than this value.
    gtol : float, default=1e-6
        Tolerance for gradient norm. Convergence occurs when the infinity
        norm of the gradient (or projected gradient for constrained problems)
        falls below this threshold.
    max_nfev : int, default=100
        Maximum number of function evaluations.
    diag : Tree, optional
        Diagonal scaling factors for variables, in the form of a tree matching
        the structure of ``x0``. If None, automatic scaling is used based on the
        Hessian diagonal. Custom scaling can improve convergence for ill-conditioned
        problems.
    lambda0 : float, default=1e-3
        Initial Levenberg-Marquardt damping parameter. Larger values
        bias toward gradient descent, smaller values toward Gauss-Newton.
    log_level : int, default=0
        Print progress every ``nprint`` iterations. Set to 0 to disable
        progress output.

    Returns
    -------
    result : OptimizeResult
        Optimization result containing:

        - ``x`` : Solution parameters with same structure as ``x0``
        - ``success`` : Whether optimization succeeded
        - ``status`` : Termination status (:class:`LMStatus`)
        - ``message`` : Descriptive termination message
        - ``fun`` : Final objective value
        - ``jac`` : Final gradient vector
        - ``hess`` : Final Hessian matrix
        - ``nfev`` : Number of function evaluations
        - ``nit`` : Number of iterations
        - ``history`` : Detailed iteration history

    Notes
    -----
    The algorithm implements the damped normal equations approach:

    .. code-block:: text

        (H + λI)p = -g

    where ``H`` is the Hessian approximation, ``λ`` is the damping parameter,
    and ``p`` is the step direction. The damping parameter is adapted based
    on the ratio of actual to predicted reduction.

    For box-constrained problems, the algorithm uses projected gradients for
    convergence testing and OSQP for constrained step computation when bounds
    are violated.  When the damped normal equations yield a step that would violate
    the bounds, the algorithm switches to solving the quadratic program:

    .. code-block:: text

        minimize    0.5 * p^T * (H + λI) * p + g^T * p
        subject to  lb <= x + p <= ub

    Examples
    --------
    >>> import numpy as np
    >>> from archimedes.sysid import lm_solve
    >>> import archimedes as arc
    >>>
    >>> # Rosenbrock function as least-squares problem
    >>> def rosenbrock_func(x):
    ...     # Residuals: r1 = 10*(x[1] - x[0]²), r2 = (1 - x[0])
    ...     return np.hstack([10*(x[1] - x[0]**2), 1 - x[0]])
    >>>
    >>> # Solve unconstrained problem
    >>> result = lm_solve(rosenbrock_func, x0=np.array([-1.2, 1.0]))
    >>> print(f"Solution: {result.x}")
    Solution: [0.99999941 0.99999881
    >>> print(f"Converged: {result.success} ({result.message})")
    Converged: True (The cosine of the angle between fvec and any column...)
    >>>
    >>> # Solve with box constraints (physical parameter limits)
    >>> bounds = (np.array([0.0, 0.0]), np.array([0.8, 2.0]))
    >>> result = lm_solve(rosenbrock_func, x0=np.array([0.5, 0.5]), bounds=bounds)
    >>> print(f"Constrained solution: {result.x}")
    Constrained solution: [0.79999998 0.6391025 ]

    See Also
    --------
    archimedes.sysid.pem : Parameter estimation using prediction error minimization
    LMStatus : Convergence status codes and meanings
    scipy.optimize.least_squares : SciPy's least-squares solver
    """
    if log_level is None:
        logger = None
    else:
        logger = logging.getLogger("levenberg_marquardt")
        logger.setLevel(log_level)

    x0_flat, bounds_flat, unravel = _ravel_args(x0, bounds)  # Validate bounds structure

    progress = LMProgress(logger)  # Initialize logger

    # Constants
    MACHEP = np.finfo(float).eps  # Machine precision
    SQRT_MACHEP = np.sqrt(MACHEP)  # Square root of machine precision

    # Initialize parameters and arrays
    x = x0_flat.copy()  # Start with the flattened initial guess
    n = len(x)

    if bounds is not None:
        lb, ub = bounds_flat  # Unravel bounds to flat arrays
        # Initialize OSQP solver for box-constrained QP
        qp_solver = osqp.OSQP()
        qp_solver.setup(
            P=sparse.csc_matrix(np.ones((n, n))),  # Will be updated each iteration
            q=np.zeros(n),  # Will be updated each iteration
            A=sparse.csc_matrix(np.eye(n)),  # Identity matrix for box constraints
            l=lb - x,  # Step lower bounds (will be updated)
            u=ub - x,  # Step upper bounds (will be updated)
            verbose=False,  # Suppress OSQP output
            eps_abs=1e-8,  # Absolute tolerance
            eps_rel=1e-8,  # Relative tolerance
        )
    else:
        qp_solver = None

    # Wrap the original function to apply the unravel and compute
    # gradient and Hessian
    _func = arc.compile(func)

    def res_func(x_flat):
        x = unravel(x_flat)
        r = _func(x, *args)
        return tree.ravel(r)[0]  # Return flattened residuals

    @arc.compile
    def func_and_grads(x):
        r = res_func(x)
        J = arc.jac(res_func)(x)
        V = 0.5 * np.sum(r**2)
        g = J.T @ r
        H = J.T @ J
        return V, g, H

    # Auto-detect scaling: if diag is None, use automatic scaling
    auto_scale = diag is None
    if diag is None:
        diag: np.ndarray = np.ones(n)  # type: ignore[no-redef]
    else:
        diag: np.ndarray = tree.ravel(diag)[0]  # type: ignore[no-redef]

    # Initialize counters and status variables
    nfev = 0  # Number of function evaluations
    njev = 0  # Number of Jacobian evaluations
    iter = 0  # Iteration counter
    status = None  # Will be set to LMStatus value

    # Always collect iteration history
    history = []

    # Initial evaluation
    cost, grad, hess = func_and_grads(x)
    nfev += 1

    # Calculate gradient norm for convergence check
    g_norm = np.linalg.norm(grad, np.inf)

    # Initialize the Levenberg-Marquardt parameter
    lambda_val = lambda0  # Initial damping parameter

    # Main iteration loop
    while nfev < max_nfev:
        # Record iteration history before computing step
        history_entry = {
            "iter": iter,
            "cost": float(cost),
            "grad_norm": float(g_norm),
            "lambda": float(lambda_val),
            "x": x.copy(),  # Current parameter values
        }

        # Add constrained optimization info if bounds are present
        if bounds_flat is not None:
            _, g_proj_norm, active_info = _check_constrained_convergence(
                grad, x, bounds_flat, gtol
            )
            history_entry.update(
                {
                    "grad_proj_norm": float(g_proj_norm),
                    "n_active_lower": int(active_info["n_active_lower"]),
                    "n_active_upper": int(active_info["n_active_upper"]),
                    "total_active": int(active_info["total_active"]),
                }
            )

        history.append(history_entry)

        # Increment Jacobian evaluations counter
        njev += 1

        # Update diagonal scaling if using automatic scaling
        if iter == 0 and auto_scale:
            # Use the diagonal of the Hessian for scaling
            diag = np.sqrt(np.maximum(np.diag(hess), 1e-8))

        # Calculate scaled vector norm
        xnorm = np.linalg.norm(diag * x)

        # Check gradient convergence (constrained vs unconstrained)
        if bounds_flat is not None:
            # Constrained optimization: use projected gradient
            converged, g_proj_norm, active_info = _check_constrained_convergence(
                grad, x, bounds_flat, gtol
            )
            if converged:
                status = LMStatus.GTOL_REACHED
                break
            # Use projected gradient norm for progress reporting
            effective_grad_norm = g_proj_norm
        else:
            # Unconstrained optimization: use standard gradient norm
            if g_norm <= gtol:
                status = LMStatus.GTOL_REACHED
                break
            effective_grad_norm = g_norm

        # Inner loop - compute step and try it
        inner_loop_exit = False
        while True:
            # Compute step using damped normal equations
            step = _compute_step(
                hess, grad, diag, lambda_val, x, bounds_flat, qp_solver
            )

            # Compute trial point
            x_new = x + step
            if logger is not None:
                logger.debug(f"lambda: {lambda_val}")
                logger.debug(f"Trial point: {x_new}")

            # Compute scaled step norm
            pnorm = np.linalg.norm(diag * step)

            # Evaluate function at trial point
            cost_new, grad_new, hess_new = func_and_grads(x_new)
            nfev += 1

            if logger is not None:
                logger.debug(
                    f"Trial func eval: cost={cost_new}, grad={grad_new}, "
                    f"hess={hess_new}"
                )

            # Compute actual reduction
            actred = -1.0
            if cost_new < cost:  # Only consider actual reduction if cost decreased
                actred = 1.0 - cost_new / cost

            # Compute predicted reduction using quadratic model
            prered = _compute_predicted_reduction(grad, step, hess, cost)

            # Compute ratio of actual to predicted reduction
            ratio = 0.0
            if prered > 0.0:  # Ensure we have a positive predicted reduction
                ratio = actred / prered

            # Update lambda based on ratio (Trust region update)
            if ratio <= 0.25:
                if actred >= 0:
                    inv_scale = 0.5
                else:
                    dirder = np.dot(grad, step)
                    inv_scale = 0.5 * dirder / (dirder + 0.5 * actred)
                    inv_scale = max(inv_scale, 0.1)
            else:
                inv_scale = 2.0
            lambda_val = lambda_val / inv_scale

            # Test for successful iteration
            if ratio >= 1.0e-4:  # Step provides sufficient decrease
                # Accept the step
                x = x_new
                cost, grad, hess = cost_new, grad_new, hess_new
                g_norm = np.linalg.norm(grad, np.inf)
                xnorm = np.linalg.norm(diag * x)

                # Update effective gradient norm for progress reporting
                if bounds_flat is not None:
                    _, g_proj_norm, _ = _check_constrained_convergence(
                        grad, x, bounds_flat, gtol
                    )
                    effective_grad_norm = g_proj_norm
                else:
                    effective_grad_norm = g_norm

                # Report progress (use appropriate gradient norm)
                progress.report(cost, effective_grad_norm, pnorm, nfev)

                # Machine precision check: if step norm is tiny relative to
                # parameter norm, we're likely at numerical precision limits
                # and should terminate
                if pnorm <= SQRT_MACHEP * max(xnorm, 1.0):
                    status = LMStatus.XTOL_REACHED
                    break

                # Record detailed step information in history
                if len(history) > 0:
                    # Update current iteration's history with step details
                    history[-1].update(
                        {
                            "step_norm": float(pnorm),
                            "actred": float(actred),
                            "prered": float(prered),
                            "ratio": float(ratio),
                            "lambda_next": float(lambda_val),
                        }
                    )

                iter += 1
                break

            # If maximum function evaluations reached during inner loop
            if nfev >= max_nfev:
                status = LMStatus.MAX_FEVAL
                inner_loop_exit = True
                break

        # Check if we exited inner loop due to max function evaluations
        if inner_loop_exit:
            break

        # Test convergence conditions
        # 1. Function value convergence (ftol)
        if abs(actred) <= ftol and prered <= ftol and 0.5 * ratio <= 1.0:
            status = LMStatus.FTOL_REACHED

        # 2. Parameter convergence (xtol)
        # Check step size relative to parameter magnitude
        if pnorm <= xtol * xnorm:
            # Check if we also satisfied ftol for combined convergence
            if status == LMStatus.FTOL_REACHED:
                status = LMStatus.BOTH_TOL_REACHED
            else:
                status = LMStatus.XTOL_REACHED

        if status is not None:
            break

    if status is None:
        status = LMStatus.MAX_FEVAL

    if logger is None:
        print(status.message)
    else:
        logger.info(status.message)

    # Calculate final residuals
    res = res_func(x)

    # Unravel the final solution
    x = unravel(x)

    # Create and return result
    return LMResult(
        x=x,
        success=status.success,
        status=status,
        message=status.message,
        fun=res,  # Final residuals value
        jac=grad,
        hess=hess,
        nfev=nfev,
        njev=njev,
        nit=iter,
        history=history,
    )



import warnings
import sys

from . import aero, coco, observers, signal
from .lqr import lqr_design
from .balanced_truncation import balanced_truncation


def __getattr__(name):
    if name == "spatial":
        warnings.warn(
            "Importing from archimedes.experimental.spatial is deprecated "
            "and will be removed in version 1.0. "
            "Please import from archimedes.spatial instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        # Import the new module
        from archimedes import spatial

        # Cache it in sys.modules under the old path
        # This ensures subsequent imports use the same module object
        sys.modules["archimedes.experimental.spatial"] = spatial

        return spatial


__all__ = [
    "coco",
    "aero",
    "observers",
    "signal",
    "spatial",
    "lqr_design",
    "balanced_truncation",
]



#
# Copyright (c) 2025 Pine Tree Labs, LLC.
#
# This file is part of Archimedes
# (see github.com/pinetreelabs/archimedes).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.#
import numpy as np
from scipy import linalg


def balanced_truncation(A, B, C, D=None, r=None, tol=1e-5):
    """
    Balanced truncation for linear time-invariant systems using NumPy/SciPy.

    Parameters:
    -----------
    A : ndarray
        System matrix, shape (n, n)
    B : ndarray
        Input matrix, shape (n, m)
    C : ndarray
        Output matrix, shape (p, n)
    D : ndarray, optional
        Feedthrough matrix, shape (p, m). Default is None.
    r : int, optional
        Order of the reduced model. Default is None (automatically determined).
    tol : float, optional
        Tolerance for automatic order selection. Default is 1e-5.
        Only used if r is None.

    Returns:
    --------
    A_r : ndarray
        Reduced system matrix, shape (r, r)
    B_r : ndarray
        Reduced input matrix, shape (r, m)
    C_r : ndarray
        Reduced output matrix, shape (p, r)
    D_r : ndarray
        Reduced feedthrough matrix, shape (p, m)
    hsv : ndarray
        Hankel singular values
    """
    # Check inputs
    A = np.asarray(A)
    B = np.asarray(B)
    C = np.asarray(C)

    if not (A.ndim == 2 and A.shape[0] == A.shape[1]):
        raise ValueError("A must be a square matrix")

    n = A.shape[0]

    if not (B.ndim == 2 and B.shape[0] == n):
        raise ValueError("B must have the same number of rows as A")

    if not (C.ndim == 2 and C.shape[1] == n):
        raise ValueError("C must have the same number of columns as A")

    # Set default D matrix if not provided
    if D is None:
        p, m = C.shape[0], B.shape[1]
        D = np.zeros((p, m))
    else:
        D = np.asarray(D)
        if not (D.ndim == 2 and D.shape[0] == C.shape[0] and D.shape[1] == B.shape[1]):
            raise ValueError(
                "D must have shape (p, m) where p is the number of rows in C and m is the number of columns in B"
            )

    # Check system stability before proceeding
    eigs = linalg.eigvals(A)
    if np.max(np.real(eigs)) >= 0:
        raise ValueError(
            "System is not stable. Balanced truncation requires a stable system."
        )

    # Solve Lyapunov equations for Gramians with improved numerical conditioning
    # For controllability Gramian: AWc + WcA^T + BB^T = 0
    Wc = linalg.solve_continuous_lyapunov(A, -B @ B.T)

    # For observability Gramian: A^TWo + WoA + C^TC = 0
    Wo = linalg.solve_continuous_lyapunov(A.T, -C.T @ C)

    # Ensure symmetry (numerical stability)
    Wc = (Wc + Wc.T) / 2
    Wo = (Wo + Wo.T) / 2

    # Check if Gramians are positive definite
    try:
        # This will either apply Cholesky or eigendecomposition based on condition
        sqrt_method = "auto"
        Lc, Lo, U, hsv, Vh = _compute_balancing_transform(Wc, Wo, sqrt_method)
    except Exception as e:
        print(f"Warning: First method failed. Trying with eigendecomposition: {e}")
        sqrt_method = "eig"
        Lc, Lo, U, hsv, Vh = _compute_balancing_transform(Wc, Wo, sqrt_method)

    # Determine reduced order
    if r is None:
        # Automatic order selection based on singular value decay
        if len(hsv) == 1:
            r = 1
        else:
            # Normalize singular values
            hsv_normalized = hsv / hsv[0]
            # Find where normalized HSV falls below tolerance
            r_indices = np.where(hsv_normalized < tol)[0]
            r = r_indices[0] if len(r_indices) > 0 else n
            # Ensure at least one state is kept
            r = max(1, r)

    r = min(r, n)  # Ensure r doesn't exceed system order

    # Truncate to desired order
    hsv_r = hsv[:r]
    U_r = U[:, :r]
    Vh_r = Vh[:r, :]

    # Compute balancing transformation and its inverse
    Sigma_r_inv_sqrt = np.diag(hsv_r ** (-0.5))
    T = Lc @ Vh_r.T @ Sigma_r_inv_sqrt  # Balancing transformation
    Tinv = Sigma_r_inv_sqrt @ U_r.T @ Lo.T  # Inverse transformation

    # Apply balancing transformation and truncate
    A_r = Tinv @ A @ T
    B_r = Tinv @ B
    C_r = C @ T
    D_r = D

    return A_r, B_r, C_r, D_r, hsv


def _compute_balancing_transform(Wc, Wo, method="auto"):
    """
    Helper function to compute the balancing transform with different methods.

    Parameters:
    -----------
    Wc : ndarray
        Controllability Gramian
    Wo : ndarray
        Observability Gramian
    method : str
        Method for computing square root: 'auto', 'chol' or 'eig'

    Returns:
    --------
    Lc : ndarray
        Cholesky/square root factor of Wc
    Lo : ndarray
        Cholesky/square root factor of Wo
    U : ndarray
        Left singular vectors
    hsv : ndarray
        Hankel singular values
    Vh : ndarray
        Right singular vectors
    """
    if method == "auto":
        # Try Cholesky first, fall back to eigendecomposition
        try:
            return _compute_balancing_transform(Wc, Wo, "chol")
        except:
            return _compute_balancing_transform(Wc, Wo, "eig")

    elif method == "chol":
        # Use Cholesky decomposition (more efficient)
        Lc = linalg.cholesky(Wc, lower=True)
        Lo = linalg.cholesky(Wo, lower=True)

        # SVD of the product Lo^T @ Lc
        U, hsv, Vh = linalg.svd(Lo.T @ Lc, full_matrices=False)
        return Lc, Lo, U, hsv, Vh

    elif method == "eig":
        # Eigendecomposition of Gramians
        evals_c, evecs_c = linalg.eigh(Wc)
        evals_o, evecs_o = linalg.eigh(Wo)

        # Filter out negative eigenvalues (numerical issues)
        evals_c = np.maximum(evals_c, 0)
        evals_o = np.maximum(evals_o, 0)

        # Compute square root factors
        Rc = evecs_c @ np.diag(np.sqrt(evals_c))
        Ro = evecs_o @ np.diag(np.sqrt(evals_o))

        # SVD of the product Ro^T @ Rc
        U, hsv, Vh = linalg.svd(Ro.T @ Rc, full_matrices=False)

        return Rc, Ro, U, hsv, Vh

    else:
        raise ValueError(f"Unknown method: {method}")



# ruff: noqa: N802, N803, N806

import numpy as np
import pytest

from archimedes.experimental.aero.atmosphere import (
    StandardAtmosphere1976,
)


class TestStandardAtmosphere1976:
    """Test suite for the StandardAtmosphere1976 class."""

    @pytest.fixture
    def atm(self) -> StandardAtmosphere1976:
        """Create an instance of StandardAtmosphere1976 for testing."""
        return StandardAtmosphere1976()

    def test_sea_level_values(self, atm: StandardAtmosphere1976):
        """Test that sea level values match standard values."""
        # Standard sea-level values
        T0 = 288.15  # K
        P0 = 101325  # Pa

        assert np.isclose(atm.calc_T(0), T0, rtol=1e-5)
        assert np.isclose(atm.calc_p(0), P0, rtol=1e-5)

    def test_temperature_gradient(self, atm: StandardAtmosphere1976):
        """Test temperature calculations at different altitudes."""
        # Test troposphere
        assert np.isclose(atm.calc_T(5000), 255.65, rtol=1e-4)

        # Test tropopause (isothermal layer)
        assert np.isclose(atm.calc_T(11000), 216.65, rtol=1e-4)
        assert np.isclose(atm.calc_T(15000), 216.65, rtol=1e-4)

        # Test stratosphere
        assert np.isclose(atm.calc_T(25000), 221.65, rtol=1e-4)

    def test_pressure_gradient(self, atm: StandardAtmosphere1976):
        """Test pressure calculations at different altitudes."""
        # Test various altitudes
        assert np.isclose(atm.calc_p(5000), 54048.2, rtol=1e-3)
        assert np.isclose(atm.calc_p(11000), 22632.1, rtol=1e-3)
        assert np.isclose(atm.calc_p(20000), 5474.9, rtol=1e-3)

    def test_density_calculation(self, atm: StandardAtmosphere1976):
        """Test density calculations at different altitudes."""
        # Standard values from tables
        altitudes = [0, 5000, 11000, 20000]
        expected_densities = [1.225, 0.7364, 0.3639, 0.0880]

        for alt, exp_density in zip(altitudes, expected_densities):
            # Calculate density from pressure and temperature
            T = atm.calc_T(alt)
            p = atm.calc_p(alt)
            R = 287.05287  # J/(kg·K)
            density = p / (R * T)
            assert np.isclose(density, exp_density, rtol=1e-3)

    def test_mach_and_qbar(self, atm: StandardAtmosphere1976):
        """Test Mach number and dynamic pressure calculations."""
        # Test at sea level with 100 m/s speed
        mach, qbar = atm(100, 0)

        # Expected values
        # Speed of sound at sea level: sqrt(1.4 * 287.05287 * 288.15) = 340.3 m/s
        exp_mach = 100 / 340.3
        exp_qbar = 0.5 * 1.225 * (100**2)

        assert np.isclose(mach, exp_mach, rtol=1e-3)
        assert np.isclose(qbar, exp_qbar, rtol=1e-3)

        # Test at 11000m with 200 m/s speed
        mach, qbar = atm(200, 11000)

        # Speed of sound at 11000m: sqrt(1.4 * 287.05287 * 216.65) = 295.1 m/s
        # Density at 11000m: 0.3639 kg/m³
        exp_mach_high = 200 / 295.1
        exp_qbar_high = 0.5 * 0.3639 * (200**2)

        assert np.isclose(mach, exp_mach_high, rtol=1e-3)
        assert np.isclose(qbar, exp_qbar_high, rtol=1e-3)

    def test_altitude_range(self, atm: StandardAtmosphere1976):
        """Test handling of altitudes outside the valid range."""
        # Test negative altitude (should use sea level values)
        assert np.isclose(atm.calc_T(-100), atm.calc_T(0), rtol=1e-5)
        assert np.isclose(atm.calc_p(-100), atm.calc_p(0), rtol=1e-5)

        # Test very high altitude (model is valid up to 86km)
        assert atm.calc_T(90000) is not None
        assert atm.calc_p(90000) is not None

    def test_array_inputs(self, atm: StandardAtmosphere1976):
        """Test that array inputs work correctly."""
        altitudes = np.array([0, 5000, 11000])
        temperatures = np.array([288.15, 255.65, 216.65])

        calc_temps = np.array([atm.calc_T(alt) for alt in altitudes])
        assert np.allclose(calc_temps, temperatures, rtol=1e-4)

    def test_reference_data(self, atm: StandardAtmosphere1976):
        """Test against official USSA1976 reference data."""
        # Example reference data (altitude in m, temperature in K, pressure in Pa)
        reference_data = [
            # Alt(m),   Temp(K),    Pressure(Pa),  Density(kg/m³)
            [0, 288.15, 101325, 1.225],
            [1000, 281.65, 89876, 1.112],
            [2000, 275.15, 79501, 1.007],
            [5000, 255.65, 54048, 0.7364],
            [10000, 223.15, 26500, 0.4135],
            [11000, 216.65, 22632, 0.3639],
            [15000, 216.65, 12112, 0.1948],
            [20000, 216.65, 5529, 0.0889],
            [25000, 221.65, 2511, 0.0395],
            [30000, 226.65, 1172, 0.0180],
            [40000, 251.05, 277.5, 0.0038],
            [50000, 270.65, 75.95, 0.001],
        ]

        for alt, temp, press, dens in reference_data:
            print(f"Testing altitude: {alt} m")
            assert np.isclose(atm.calc_T(alt), temp, rtol=1e-3)
            assert np.isclose(atm.calc_p(alt), press, rtol=1e-2)

            # Test Mach and qbar at 100 m/s
            mach, qbar = atm(100, alt)
            speed_of_sound = np.sqrt(1.4 * 287.05287 * temp)
            exp_mach = 100 / speed_of_sound
            exp_qbar = 0.5 * dens * (100**2)

            assert np.isclose(mach, exp_mach, rtol=1e-3)
            assert np.isclose(qbar, exp_qbar, rtol=1e-1)
